from __future__ import annotations

from contextlib import ExitStack
from dataclasses import dataclass, field, replace
from itertools import chain
from typing import TYPE_CHECKING, Any, Self

from utilities.ipython import check_ipython_class

if TYPE_CHECKING:
    from types import TracebackType


def is_jupyter() -> bool:
    """Check if `jupyter` is running."""
    try:
        from ipykernel.zmqshell import (  # pyright: ignore[reportMissingImports]
            ZMQInteractiveShell,
        )
    except ImportError:
        return False
    return check_ipython_class(ZMQInteractiveShell)  # pragma: no cover


_DEFAULT_ROWS = 7
_DEFAULT_COLS = 100


@dataclass
class _Show:
    """Context manager which adjusts the display of NDFrames."""

    rows: int | None = field(default=_DEFAULT_ROWS)
    columns: int | None = field(default=_DEFAULT_COLS, kw_only=True)
    dp: int | None = field(default=None, kw_only=True)
    stack: ExitStack = field(default_factory=ExitStack, kw_only=True)

    def __call__(
        self,
        rows: int | None = _DEFAULT_ROWS,
        *,
        dp: int | None = None,
        columns: int | None = _DEFAULT_COLS,
    ) -> Self:
        return replace(self, rows=rows, dp=dp, columns=columns)

    def __enter__(self) -> None:
        self._enter_pandas()
        self._enter_polars()
        _ = self.stack.__enter__()

    def _enter_pandas(self) -> None:
        try:
            from pandas import option_context
        except ModuleNotFoundError:  # pragma: no cover
            pass
        else:
            kwargs: dict[str, Any] = {}
            if self.dp is not None:
                kwargs["display.precision"] = self.dp
            if self.rows is not None:
                kwargs["display.min_rows"] = kwargs["display.max_rows"] = self.rows
            if self.columns is not None:
                kwargs["display.max_columns"] = self.columns
            if len(kwargs) >= 1:
                context = option_context(*chain(*kwargs.items()))
                self.stack.enter_context(context)

    def _enter_polars(self) -> None:
        try:
            from polars import Config
        except ModuleNotFoundError:  # pragma: no cover
            pass
        else:
            kwargs: dict[str, Any] = {}
            if self.dp is not None:
                kwargs["float_precision"] = self.dp
            if self.rows is not None:
                kwargs["tbl_rows"] = self.rows
            if self.columns is not None:
                kwargs["tbl_cols"] = self.columns
            config = Config(**kwargs)
            _ = self.stack.enter_context(config)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        _ = self.stack.__exit__(exc_type, exc_val, exc_tb)


show = _Show()


__all__ = ["is_jupyter", "show"]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeGuard

from pytest import mark, param, raises

from utilities.more_itertools import (
    ResolveIncludeAndExcludeError,
    Split,
    filter_include_and_exclude,
    partition_typeguard,
    peekable,
    resolve_include_and_exclude,
    yield_splits,
)
from utilities.text import strip_and_dedent

if TYPE_CHECKING:
    from collections.abc import Iterable


class TestFilterIncludeAndExclude:
    def test_none(self) -> None:
        rng = list(range(5))
        result = list(filter_include_and_exclude(rng))
        assert result == rng

    def test_include_singleton(self) -> None:
        result = list(filter_include_and_exclude(range(5), include=0))
        expected = [0]
        assert result == expected

    def test_include_iterable(self) -> None:
        result = list(filter_include_and_exclude(range(5), include=[0, 1, 2]))
        expected = [0, 1, 2]
        assert result == expected

    def test_exclude_singleton(self) -> None:
        result = list(filter_include_and_exclude(range(5), exclude=0))
        expected = [1, 2, 3, 4]
        assert result == expected

    def test_exclude_iterable(self) -> None:
        result = list(filter_include_and_exclude(range(5), exclude=[0, 1, 2]))
        expected = [3, 4]
        assert result == expected

    def test_both(self) -> None:
        result = list(
            filter_include_and_exclude(range(5), include=[0, 1], exclude=[3, 4])
        )
        expected = [0, 1]
        assert result == expected

    def test_include_key(self) -> None:
        @dataclass(frozen=True, kw_only=True)
        class Example:
            n: int

        result = list(
            filter_include_and_exclude(
                [Example(n=n) for n in range(5)], include=[0, 1, 2], key=lambda x: x.n
            )
        )
        expected = [Example(n=n) for n in [0, 1, 2]]
        assert result == expected

    def test_exclude_key(self) -> None:
        @dataclass(frozen=True, kw_only=True)
        class Example:
            n: int

        result = list(
            filter_include_and_exclude(
                [Example(n=n) for n in range(5)], exclude=[0, 1, 2], key=lambda x: x.n
            )
        )
        expected = [Example(n=n) for n in [3, 4]]
        assert result == expected


class TestPartitionTypeguard:
    def test_main(self) -> None:
        def is_int(x: Any, /) -> TypeGuard[int]:
            return isinstance(x, int)

        iterable = [1, 2.0, 3, 4.0]
        false, true = partition_typeguard(is_int, iterable)
        for el in false:
            assert isinstance(el, int | float)
        for el in true:
            assert isinstance(el, int)


class TestPeekable:
    def test_dropwhile(self) -> None:
        it = peekable(range(10))
        it.dropwhile(lambda x: x <= 4)
        assert it.peek() == 5
        result = list(it)
        expected = [5, 6, 7, 8, 9]
        assert result == expected

    def test_next(self) -> None:
        it = peekable(range(10))
        value = next(it)
        assert isinstance(value, int)

    def test_peek_non_empty(self) -> None:
        it = peekable(range(10))
        value = it.peek()
        assert isinstance(value, int)

    def test_peek_empty_without_default(self) -> None:
        it: peekable[int] = peekable([])
        with raises(StopIteration):
            _ = it.peek()

    def test_peek_empty_with_default(self) -> None:
        it: peekable[int] = peekable([])
        value = it.peek(default="default")
        assert isinstance(value, str)

    def test_takewhile(self) -> None:
        it = peekable(range(10))
        result1 = list(it.takewhile(lambda x: x <= 4))
        expected1 = [0, 1, 2, 3, 4]
        assert result1 == expected1
        assert it.peek() == 5
        result2 = list(it)
        expected2 = [5, 6, 7, 8, 9]
        assert result2 == expected2

    def test_combined(self) -> None:
        it = peekable(range(10))
        result1 = list(it.takewhile(lambda x: x <= 2))
        expected1 = [0, 1, 2]
        assert result1 == expected1
        assert it.peek() == 3
        it.dropwhile(lambda x: x <= 4)
        assert it.peek() == 5
        result2 = list(it.takewhile(lambda x: x <= 6))
        expected2 = [5, 6]
        assert result2 == expected2
        result3 = list(it)
        expected3 = [7, 8, 9]
        assert result3 == expected3


class TestResolveIncludeAndExclude:
    def test_none(self) -> None:
        include, exclude = resolve_include_and_exclude()
        assert include is None
        assert exclude is None

    def test_include_singleton(self) -> None:
        include, exclude = resolve_include_and_exclude(include=1)
        assert include == {1}
        assert exclude is None

    def test_include_iterable(self) -> None:
        include, exclude = resolve_include_and_exclude(include=[1, 2, 3])
        assert include == {1, 2, 3}
        assert exclude is None

    def test_exclude_singleton(self) -> None:
        include, exclude = resolve_include_and_exclude(exclude=1)
        assert include is None
        assert exclude == {1}

    def test_exclude_iterable(self) -> None:
        include, exclude = resolve_include_and_exclude(exclude=[1, 2, 3])
        assert include is None
        assert exclude == {1, 2, 3}

    def test_both(self) -> None:
        include, exclude = resolve_include_and_exclude(
            include=[1, 2, 3], exclude=[4, 5, 6]
        )
        assert include == {1, 2, 3}
        assert exclude == {4, 5, 6}

    def test_error(self) -> None:
        with raises(
            ResolveIncludeAndExcludeError,
            match="Iterables .* and .* must not overlap; got .*",
        ):
            _ = resolve_include_and_exclude(include=[1, 2, 3], exclude=[3, 4, 5])


class TestYieldSplits:
    @mark.parametrize(
        ("iterable", "head", "tail", "min_frac", "freq", "expected"),
        [
            param(
                "abcde",
                3,
                1,
                None,
                None,
                [
                    Split(head=["a", "b", "c"], tail=["d"]),
                    Split(head=["b", "c", "d"], tail=["e"]),
                ],
                id="3/1",
            ),
            param(
                "abcde",
                3,
                1,
                0.4,
                None,
                [
                    Split(head=["a", "b"], tail=["c"]),
                    Split(head=["a", "b", "c"], tail=["d"]),
                    Split(head=["b", "c", "d"], tail=["e"]),
                ],
                id="3/1, min-frac=0.4",
            ),
            param(
                "abcdefg",
                3,
                2,
                None,
                None,
                [
                    Split(head=["a", "b", "c"], tail=["d", "e"]),
                    Split(head=["c", "d", "e"], tail=["f", "g"]),
                ],
                id="3/2, clean tail",
            ),
            param(
                "abcdefgh",
                3,
                2,
                None,
                None,
                [
                    Split(head=["a", "b", "c"], tail=["d", "e"]),
                    Split(head=["c", "d", "e"], tail=["f", "g"]),
                    Split(head=["e", "f", "g"], tail=["h"]),
                ],
                id="3/2, truncated tail",
            ),
            param(
                "abcdefgh",
                3,
                2,
                None,
                1,
                [
                    Split(head=["a", "b", "c"], tail=["d", "e"]),
                    Split(head=["b", "c", "d"], tail=["e", "f"]),
                    Split(head=["c", "d", "e"], tail=["f", "g"]),
                    Split(head=["d", "e", "f"], tail=["g", "h"]),
                    Split(head=["e", "f", "g"], tail=["h"]),
                ],
                id="3/2, freq=1",
            ),
            param("abc", 5, 1, None, None, [], id="len(iterable) < head"),
            param("abc", 1, 5, None, None, [], id="len(iterable) < tail"),
        ],
    )
    def test_main(
        self,
        *,
        iterable: Iterable[str],
        head: int,
        tail: int,
        min_frac: float | None,
        freq: int | None,
        expected: list[Split[list[str]]],
    ) -> None:
        splits = list(yield_splits(iterable, head, tail, min_frac=min_frac, freq=freq))
        assert splits == expected

    def test_repr(self) -> None:
        split = Split(head=["a", "b", "c"], tail=["d"])
        result = repr(split)
        expected = strip_and_dedent(
            """
            Split(
                head=
                    ['a', 'b', 'c']
                tail=
                    ['d']
            )
            """
        )
        assert result == expected



from __future__ import annotations

from asyncio import sleep
from dataclasses import dataclass
from functools import partial
from itertools import chain, repeat
from typing import TYPE_CHECKING, Any, ClassVar

from hypothesis import given
from pytest import mark, param, raises

from utilities.asyncio import (
    ReduceAsyncError,
    _MaybeAwaitableMaybeAsyncIterable,
    groupby_async,
    groupby_async_list,
    is_awaitable,
    reduce_async,
    timeout_dur,
    to_list,
    to_set,
    to_sorted,
    try_await,
)
from utilities.hypothesis import durations

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterable, Iterator, Sequence

    from utilities.types import Duration

_STRS = list("AAAABBBCCDAABB")


def _get_strs_sync() -> Iterable[str]:
    return iter(_STRS)


async def _get_strs_async() -> Iterable[str]:
    return _get_strs_sync()


def _yield_strs_sync() -> Iterator[str]:
    return iter(_get_strs_sync())


async def _yield_strs_async() -> AsyncIterator[str]:
    for i in _get_strs_sync():
        yield i
        await sleep(0.01)


@dataclass(frozen=True, kw_only=True)
class _Container:
    text: str


def _get_containers_sync() -> Iterable[_Container]:
    return (_Container(text=t) for t in _get_strs_sync())


async def _get_containers_async() -> Iterable[_Container]:
    return _get_containers_sync()


def _yield_containers_sync() -> Iterator[_Container]:
    return iter(_get_containers_sync())


async def _yield_containers_async() -> AsyncIterator[_Container]:
    for i in _get_containers_sync():
        yield i
        await sleep(0.01)


async def _ord_async(text: str, /) -> int:
    await sleep(0.01)
    return ord(text)


class TestGroupbyAsync:
    exp_no_key: ClassVar[list[tuple[str, list[str]]]] = [
        ("A", list(repeat("A", times=4))),
        ("B", list(repeat("B", times=3))),
        ("C", list(repeat("C", times=2))),
        ("D", list(repeat("D", times=1))),
        ("A", list(repeat("A", times=2))),
        ("B", list(repeat("B", times=2))),
    ]
    exp_with_key: ClassVar[list[tuple[int, list[str]]]] = [
        (65, list(repeat("A", times=4))),
        (66, list(repeat("B", times=3))),
        (67, list(repeat("C", times=2))),
        (68, list(repeat("D", times=1))),
        (65, list(repeat("A", times=2))),
        (66, list(repeat("B", times=2))),
    ]

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_no_key(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = groupby_async(iterable)
        as_list: list[tuple[str, list[str]]] = []
        async for k, v in await result:
            assert isinstance(k, str)
            assert isinstance(v, list)
            for v_i in v:
                assert isinstance(v_i, str)
            as_list.append((k, v))
        assert as_list == self.exp_no_key

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_no_key_as_list(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await groupby_async_list(iterable)
        assert result == self.exp_no_key

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_key_sync(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = groupby_async(iterable, key=ord)
        as_list: list[tuple[int, list[str]]] = []
        async for k, v in await result:
            assert isinstance(k, int)
            assert isinstance(v, list)
            assert all(isinstance(v_i, str) for v_i in v)
            as_list.append((k, v))
        assert as_list == self.exp_with_key

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_key_sync_list(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await groupby_async_list(iterable, key=ord)
        assert result == self.exp_with_key

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_key_async(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = groupby_async(iterable, key=_ord_async)
        as_list: list[tuple[int, list[str]]] = []
        async for k, v in await result:
            assert isinstance(k, int)
            assert isinstance(v, list)
            assert all(isinstance(v_i, str) for v_i in v)
            as_list.append((k, v))
        assert as_list == self.exp_with_key

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_key_async_list(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await groupby_async_list(iterable, key=_ord_async)
        assert result == self.exp_with_key


class TestIsAwaitable:
    @mark.parametrize(
        ("obj", "expected"), [param(sleep(0.01), True), param(None, False)]
    )
    async def test_main(self, *, obj: Any, expected: bool) -> None:
        result = await is_awaitable(obj)
        assert result is expected


class TestReduceAsync:
    async def test_no_initial(self) -> None:
        async def add(x: int, y: int, /) -> int:
            await sleep(0.01)
            return x + y

        result = await reduce_async(add, [1, 2, 3])
        assert result == 6

    async def test_no_initial_with_partial(self) -> None:
        async def add(x: int, y: int, /, *, z: int) -> int:
            await sleep(0.01)
            return x + y + z

        result = await reduce_async(partial(add, z=1), [1, 2, 3])
        assert result == 8

    async def test_with_initial(self) -> None:
        async def collect(x: Iterable[int], y: int, /) -> Sequence[int]:
            await sleep(0.01)
            return list(chain(x, [y]))

        result = await reduce_async(collect, [1, 2, 3], initial=[])
        assert result == [1, 2, 3]

    async def test_with_initial_with_partial(self) -> None:
        async def collect(x: Iterable[int], y: int, /, *, z: int) -> Sequence[int]:
            await sleep(0.01)
            return list(chain(x, [y, z]))

        result = await reduce_async(partial(collect, z=0), [1, 2, 3], initial=[])
        assert result == [1, 0, 2, 0, 3, 0]

    async def test_empty(self) -> None:
        async def add(x: int, y: int, /) -> int:
            await sleep(0.01)
            return x + y

        with raises(ReduceAsyncError, match="Empty iterable .* with no initial value"):
            _ = await reduce_async(add, [])


class TestTimeoutDur:
    @given(duration=durations())
    async def test_main(self, *, duration: Duration) -> None:
        async with timeout_dur(duration=duration):
            pass


class TestToList:
    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_main(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await to_list(iterable)
        assert result == _STRS


class TestToSet:
    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_main(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await to_set(iterable)
        assert result == set(_STRS)


class TestToSorted:
    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_main(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await to_sorted(iterable)
        expected = sorted(_STRS)
        assert result == expected

    @mark.parametrize(
        "iterable",
        [
            param(_get_containers_sync()),
            param(_get_containers_async()),
            param(_yield_containers_sync()),
            param(_yield_containers_async()),
        ],
    )
    async def test_key_sync(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[_Container]
    ) -> None:
        result = await to_sorted(iterable, key=lambda c: c.text)
        expected = [_Container(text=t) for t in sorted(_STRS)]
        assert result == expected

    @mark.parametrize(
        "iterable",
        [
            param(_get_containers_sync()),
            param(_get_containers_async()),
            param(_yield_containers_sync()),
            param(_yield_containers_async()),
        ],
    )
    async def test_key_async(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[_Container]
    ) -> None:
        async def key(container: _Container, /) -> str:
            await sleep(0.01)
            return container.text

        result = await to_sorted(iterable, key=key)
        expected = [_Container(text=t) for t in sorted(_STRS)]
        assert result == expected

    @mark.parametrize(
        "iterable",
        [
            param(_get_strs_sync()),
            param(_get_strs_async()),
            param(_yield_strs_sync()),
            param(_yield_strs_async()),
        ],
    )
    async def test_reverse(
        self, *, iterable: _MaybeAwaitableMaybeAsyncIterable[str]
    ) -> None:
        result = await to_sorted(iterable, reverse=True)
        expected = sorted(_STRS, reverse=True)
        assert result == expected


class TestTryAwait:
    async def test_maybe_awaitable(self) -> None:
        from utilities.random import SYSTEM_RANDOM

        if SYSTEM_RANDOM.random() <= 0.5:

            async def func(*, value: bool) -> bool:  # pyright: ignore[reportRedeclaration]
                await sleep(0.01)
                return not value

        else:

            def func(*, value: bool) -> bool:
                return not value

        result = await try_await(func(value=True))
        assert result is False

    async def test_awaitable(self) -> None:
        async def func(*, value: bool) -> bool:
            await sleep(0.01)
            return not value

        result = await try_await(func(value=True))
        assert result is False

    async def test_sync(self) -> None:
        def func(*, value: bool) -> bool:
            return not value

        result = await try_await(func(value=True))
        assert result is False



from __future__ import annotations

from asyncio import sleep

from utilities.loguru import LogLevel, log_call

# test sync


@log_call
def add_sync(x: int, y: int, /) -> int:
    return x + y


@log_call
def diff_sync(x: int, y: int, /) -> int:
    return x - y


@log_call
def diff_pairwise_then_add_sync(x: int, y: int, z: int, w: int, /) -> int:
    first = diff_sync(x, y)
    second = diff_sync(z, w)
    return add_sync(first, second)


# test async


@log_call
async def add_async(x: int, y: int, /) -> int:
    await sleep(0.01)
    return x + y


@log_call
async def diff_async(x: int, y: int, /) -> int:
    await sleep(0.01)
    return x - y


@log_call
async def diff_pairwise_then_add_async(x: int, y: int, z: int, w: int, /) -> int:
    first = await diff_async(x, y)
    second = await diff_async(z, w)
    return await add_async(first, second)


# test custom level


@log_call(level=LogLevel.INFO)
def add_sync_info(x: int, y: int, /) -> int:
    return x + y



from __future__ import annotations

from enum import Enum, StrEnum, auto
from itertools import chain
from typing import TYPE_CHECKING, Any

from polars import int_range
from pytest import mark, param

from utilities.reprlib import (
    ReprLocals,
    _custom_mapping_repr,
    _filter_mapping,
    custom_print,
    custom_repr,
)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


class TestCustomPrint:
    def test_main(self) -> None:
        custom_print({})


class TestCustomRepr:
    @mark.parametrize(
        ("mapping", "expected"),
        [
            param({}, ""),
            param({"a": 1}, "a=1"),
            param({"a": 1, "b": 2}, "a=1, b=2"),
            param({"a": 1, "b": 2, "c": 3}, "a=1, b=2, c=3"),
            param({"a": 1, "b": 2, "c": 3, "d": 4}, "a=1, b=2, c=3, d=4"),
            param({"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}, "a=1, b=2, c=3, d=4, ..."),
            param(
                {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
                "a=1, b=2, c=3, d=4, ...",
            ),
        ],
    )
    def test_main(self, *, mapping: Mapping[str, Any], expected: str) -> None:
        result = custom_repr(mapping)
        assert result == expected

    def test_dataframe(self) -> None:
        df = int_range(start=0, end=100, eager=True).rename("int").to_frame()
        result = custom_repr(df)
        expected = repr(df)
        assert result == expected

    def test_dataframe_fake(self) -> None:
        class DataFrame: ...

        _ = custom_repr(DataFrame())

    def test_enum_generic(self) -> None:
        class Truth(Enum):
            true = auto()
            false = auto()

        result = custom_repr(list(Truth))
        expected = "['Truth.true', 'Truth.false']"
        assert result == expected

    def test_enum_str(self) -> None:
        class Truth(StrEnum):
            true_key = "true_value"
            false_key = "false_value"

        result = custom_repr(list(Truth))
        expected = "['true_value', 'false_value']"
        assert result == expected

    def test_series(self) -> None:
        sr = int_range(start=0, end=100, eager=True).rename("int")
        result = custom_repr(sr)
        expected = repr(sr)
        assert result == expected

    def test_series_fake(self) -> None:
        class Series: ...

        _ = custom_repr(Series())


class TestCustomMappingRepr:
    @mark.parametrize(
        ("mapping", "expected"),
        [
            param({}, ""),
            param({"a": 1}, "a=1"),
            param({"a": 1, "b": 2}, "a=1, b=2"),
            param({"a": 1, "b": 2, "c": 3}, "a=1, b=2, c=3"),
            param({"a": 1, "b": 2, "c": 3, "d": 4}, "a=1, b=2, c=3, d=4"),
            param({"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}, "a=1, b=2, c=3, d=4, e=5"),
            param(
                {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6},
                "a=1, b=2, c=3, d=4, e=5, f=6",
            ),
            param({"a": [1, 2, 3, 4, 5]}, "a=[1, 2, 3, 4, 5]"),
            param({"a": [1, 2, 3, 4, 5, 6]}, "a=[1, 2, 3, 4, 5, 6]"),
            param({"a": [1, 2, 3, 4, 5, 6, 7]}, "a=[1, 2, 3, 4, 5, 6, ...]"),
            param({"a": [1, 2, 3, 4, 5, 6, 7, 8]}, "a=[1, 2, 3, 4, 5, 6, ...]"),
        ],
    )
    def test_main(self, *, mapping: Mapping[str, Any], expected: str) -> None:
        result = _custom_mapping_repr(mapping)
        assert result == expected


class TestFilterMapping:
    @mark.parametrize(
        ("include_underscore", "include_none", "expected"),
        [
            param(False, False, {"a": 1, "c": 3}),
            param(False, True, {"a": 1, "b": None, "c": 3}),
            param(True, False, {"a": 1, "c": 3, "_underscore": 4}),
            param(True, True, {"a": 1, "b": None, "c": 3, "_underscore": 4}),
        ],
    )
    def test_main(
        self,
        *,
        include_underscore: bool,
        include_none: bool,
        expected: Mapping[str, Any],
    ) -> None:
        mapping = {"a": 1, "b": None, "c": 3, "_underscore": 4}
        result = _filter_mapping(
            mapping, include_underscore=include_underscore, include_none=include_none
        )
        assert result == expected


class TestReprLocals:
    @mark.parametrize(
        ("b", "include_none", "expected"),
        [
            param(2, False, "a=1, b=2, total=3"),
            param(2, True, "a=1, b=2, total=3"),
            param(None, False, "a=1, total=1"),
            param(None, True, "a=1, b=None, total=1"),
        ],
    )
    def test_main(self, *, b: int | None, include_none: bool, expected: str) -> None:
        def func(a: int, /, *, b: int | None = None) -> str:
            init = ReprLocals(locals(), func, include_none=include_none)
            total = a if b is None else (a + b)
            return f"{init}, total={total}"

        result = func(1, b=b)
        assert result == expected

    def test_fill(self) -> None:
        def func(
            *,
            a: Sequence[int],
            b: Sequence[int],
            c: Sequence[int],
            d: Sequence[int],
            e: Sequence[int],
            f: Sequence[int],
        ) -> str:
            init = ReprLocals(locals(), func)
            total = sum(chain(a, b, c, d, e, f))
            return f"{init}, total={total}"

        eight = [1, 2, 3, 4, 5, 6, 7, 8]
        result = func(a=eight, b=eight, c=eight, d=eight, e=eight, f=eight)
        expected = "a=[1, 2, 3, 4, 5, 6, ...], b=[1, 2, 3, 4, 5, 6, ...], c=[1, 2, 3, 4, 5, 6, ...], d=[1, 2, 3, 4, 5, 6, ...], e=[1, 2, 3, 4, 5, 6, ...], f=[1, 2, 3, 4, 5, 6, ...], total=216"
        assert result == expected



from __future__ import annotations

from asyncio import timeout
from collections.abc import (
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Iterable,
    Sequence,
)
from dataclasses import dataclass
from itertools import groupby
from typing import TYPE_CHECKING, Any, TypeGuard, TypeVar, cast, overload

from typing_extensions import override

from utilities.datetime import duration_to_float
from utilities.sentinel import Sentinel, sentinel
from utilities.typing import SupportsRichComparison

if TYPE_CHECKING:
    from asyncio import Timeout

    from utilities.types import Duration


_T = TypeVar("_T")
_U = TypeVar("_U")
Coroutine1 = Coroutine[Any, Any, _T]
MaybeAwaitable = _T | Awaitable[_T]
MaybeCoroutine1 = _T | Coroutine1[_T]
_MaybeAsyncIterable = Iterable[_T] | AsyncIterable[_T]
_MaybeAwaitableMaybeAsyncIterable = MaybeAwaitable[_MaybeAsyncIterable[_T]]
_TSupportsRichComparison = TypeVar(
    "_TSupportsRichComparison", bound=SupportsRichComparison
)


@overload
async def groupby_async(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T], /, *, key: None = None
) -> AsyncIterator[tuple[_T, Sequence[_T]]]: ...
@overload
async def groupby_async(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T],
    /,
    *,
    key: Callable[[_T], MaybeAwaitable[_U]],
) -> AsyncIterator[tuple[_U, Sequence[_T]]]: ...
async def groupby_async(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T],
    /,
    *,
    key: Callable[[_T], MaybeAwaitable[_U]] | None = None,
) -> AsyncIterator[tuple[_T, Sequence[_T]]] | AsyncIterator[tuple[_U, Sequence[_T]]]:
    """Yield consecutive keys and groups (as lists)."""
    as_list = await to_list(iterable)
    if key is None:

        async def iterator1() -> AsyncIterator[tuple[_T, Sequence[_T]]]:
            for k, group in groupby(as_list):
                yield k, list(group)

        return iterator1()

    async def iterator2() -> AsyncIterator[tuple[_U, Sequence[_T]]]:
        keys = [await try_await(key(e)) for e in as_list]
        pairs = zip(keys, as_list, strict=True)
        for k, pairs_group in groupby(pairs, key=lambda x: x[0]):
            group = [v for _, v in pairs_group]
            yield k, group

    return iterator2()


@overload
async def groupby_async_list(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T], /, *, key: None = None
) -> Sequence[tuple[_T, Sequence[_T]]]: ...
@overload
async def groupby_async_list(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T],
    /,
    *,
    key: Callable[[_T], MaybeAwaitable[_U]],
) -> Sequence[tuple[_U, Sequence[_T]]]: ...
async def groupby_async_list(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T],
    /,
    *,
    key: Callable[[_T], MaybeAwaitable[_U]] | None = None,
) -> Sequence[tuple[_T, Sequence[_T]]] | Sequence[tuple[_U, Sequence[_T]]]:
    """Yield consecutive keys and groups (as lists)."""
    if key is None:
        iterator = await groupby_async(iterable)
        return await to_list(iterator)
    iterator = await groupby_async(iterable, key=key)
    return await to_list(iterator)


async def is_awaitable(obj: Any, /) -> TypeGuard[Awaitable[Any]]:
    """Check if an object is awaitable."""
    try:
        await obj
    except TypeError:
        return False
    return True


@overload
async def reduce_async(
    func: Callable[[_T, _U], Awaitable[_T]], iterable: Iterable[_U], /, *, initial: _T
) -> _T: ...
@overload
async def reduce_async(
    func: Callable[[_T, _T], Awaitable[_T]],
    iterable: Iterable[_T],
    /,
    *,
    initial: Sentinel = sentinel,
) -> _T: ...
async def reduce_async(
    func: Callable[[Any, Any], Awaitable[Any]],
    iterable: Iterable[Any],
    /,
    *,
    initial: Any = sentinel,
) -> Any:
    """Apply a function of two arguments cumulatively to an iterable."""
    if isinstance(initial, Sentinel):
        iterator = iter(iterable)
        try:
            value = next(iterator)
        except StopIteration:
            raise ReduceAsyncError(
                func=func, iterable=iterable, initial=initial
            ) from None
    else:
        iterator = iterable
        value = initial
    for element in iterator:
        value = await func(value, element)
    return value


@dataclass(kw_only=True)
class ReduceAsyncError(Exception):
    func: Callable[[Any, Any], Awaitable[Any]]
    iterable: Iterable[Any]
    initial: Any = sentinel

    @override
    def __str__(self) -> str:
        return f"Empty iterable {self.iterable} with no initial value"


def timeout_dur(*, duration: Duration | None = None) -> Timeout:
    """Timeout context manager which accepts durations."""
    delay = None if duration is None else duration_to_float(duration)
    return timeout(delay)


async def to_list(iterable: _MaybeAwaitableMaybeAsyncIterable[_T], /) -> list[_T]:
    """Reify an asynchronous iterable as a list."""
    value = await try_await(iterable)
    try:
        return [x async for x in cast(AsyncIterable[_T], value)]
    except TypeError:
        return list(cast(Iterable[_T], value))


async def to_set(iterable: _MaybeAwaitableMaybeAsyncIterable[_T], /) -> set[_T]:
    """Reify an asynchronous iterable as a set."""
    value = await try_await(iterable)
    try:
        return {x async for x in cast(AsyncIterable[_T], value)}
    except TypeError:
        return set(cast(Iterable[_T], value))


@overload
async def to_sorted(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_TSupportsRichComparison],
    /,
    *,
    key: None = None,
    reverse: bool = ...,
) -> list[_TSupportsRichComparison]: ...
@overload
async def to_sorted(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T],
    /,
    *,
    key: Callable[[_T], MaybeAwaitable[SupportsRichComparison]],
    reverse: bool = ...,
) -> list[_T]: ...
async def to_sorted(
    iterable: _MaybeAwaitableMaybeAsyncIterable[_T]
    | _MaybeAwaitableMaybeAsyncIterable[_TSupportsRichComparison],
    /,
    *,
    key: Callable[[_T], MaybeAwaitable[SupportsRichComparison]] | None = None,
    reverse: bool = False,
) -> list[_T] | list[_TSupportsRichComparison]:
    """Convert."""
    as_list = await to_list(iterable)
    if key is None:
        as_list = cast(list[_TSupportsRichComparison], as_list)
        return sorted(as_list, reverse=reverse)

    as_list = cast(list[_T], as_list)
    values = [await try_await(key(e)) for e in as_list]
    sorted_pairs = sorted(zip(as_list, values, strict=True), key=lambda x: x[1])
    return [element for element, _ in sorted_pairs]


async def try_await(obj: MaybeAwaitable[_T], /) -> _T:
    """Try await a value from an object."""
    try:
        return await cast(Awaitable[_T], obj)
    except TypeError:
        return cast(_T, obj)


__all__ = [
    "Coroutine1",
    "MaybeAwaitable",
    "MaybeCoroutine1",
    "ReduceAsyncError",
    "groupby_async",
    "groupby_async_list",
    "is_awaitable",
    "reduce_async",
    "timeout_dur",
    "to_list",
    "to_set",
    "to_sorted",
    "try_await",
]



from __future__ import annotations

import gzip
from pickle import dump, load
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from utilities.types import PathLike


def read_pickle(path: PathLike, /) -> Any:
    """Read an object from disk."""
    with gzip.open(path, mode="rb") as gz:
        return load(gz)  # noqa: S301


def write_pickle(obj: Any, path: PathLike, /, *, overwrite: bool = False) -> None:
    """Write an object to disk."""
    from utilities.atomicwrites import writer

    with writer(path, overwrite=overwrite) as temp, gzip.open(temp, mode="wb") as gz:
        dump(obj, gz)


__all__ = ["read_pickle", "write_pickle"]
